# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/001_ConsumerLoop.ipynb.

# %% auto 0
__all__ = ['logger', 'process_msgs', 'process_message_callback', 'aiokafka_consumer_loop']

# %% ../../nbs/001_ConsumerLoop.ipynb 1
from typing import *

from os import environ
import asyncio
from datetime import datetime, timedelta
from asyncio import iscoroutinefunction  # do not use the version from inspect

from aiokafka import AIOKafkaConsumer
from aiokafka.structs import TopicPartition, ConsumerRecord
from pydantic import BaseModel, HttpUrl, NonNegativeInt, Field
import asyncer
import anyio

from .logger import get_logger

# %% ../../nbs/001_ConsumerLoop.ipynb 6
logger = get_logger(__name__)

# %% ../../nbs/001_ConsumerLoop.ipynb 10
async def process_msgs(
    *,
    msgs: Dict[TopicPartition, List[ConsumerRecord]],
    callbacks: Dict[str, Callable[[BaseModel], None]],
    msg_types: Dict[str, Type[BaseModel]],
    process_f: Callable[[Callable[[BaseModel], None], BaseModel], None]
) -> None:
    """For each messages **msg** in **msgs**, calls process_f with callbacks[topic] and **msgs**.

    Params:
        msgs: a dictionary mapping topic partition to a list of messages, returned by `AIOKafkaConsumer.getmany`.
        callbacks: a dictionary mapping topics into a callback functions.
        msg_types: a dictionary mapping topics into a message type of a message.
        process_f: a stream processing function registrated by `anyio.create_memory_object_stream`

    Todo:
        remove it :)
    """
    for topic_partition, topic_msgs in msgs.items():
        topic = topic_partition.topic
        msg_type = msg_types[topic]
        decoded_msgs = [
            msg_type.parse_raw(msg.value.decode("utf-8")) for msg in topic_msgs
        ]
        for msg in decoded_msgs:
            callback = callbacks[topic]
            if not iscoroutinefunction(callback):
                callback = asyncer.asyncify(callback)
            await process_f((callback, msg))

# %% ../../nbs/001_ConsumerLoop.ipynb 16
async def process_message_callback(receive_stream):
    async with receive_stream:
        async for callback, msg in receive_stream:
            await callback(msg)


async def _aiokafka_consumer_loop(
    consumer: AIOKafkaConsumer,
    *,
    callbacks: Dict[str, Callable[[BaseModel], None]],
    timeout_ms: int = 100,
    max_buffer_size: int = 10_000,
    msg_types: Dict[str, Type[BaseModel]],
    is_shutting_down_f: Callable[[], bool],
):
    """Write docs

    Todo: add batch size if needed
    """
    send_stream, receive_stream = anyio.create_memory_object_stream(
        max_buffer_size=max_buffer_size
    )
    async with anyio.create_task_group() as tg:
        tg.start_soon(process_message_callback, receive_stream)
        async with send_stream:
            while not is_shutting_down_f():
                msgs = await consumer.getmany(timeout_ms=timeout_ms)
                await process_msgs(
                    msgs=msgs,
                    callbacks=callbacks,
                    msg_types=msg_types,
                    process_f=send_stream.send,
                )

# %% ../../nbs/001_ConsumerLoop.ipynb 18
async def aiokafka_consumer_loop(
    topics: List[str],
    *,
    bootstrap_servers: str,
    auto_offset_reset: str,
    max_poll_records: int = 1_000,
    timeout_ms: int = 100,
    max_buffer_size: int = 10_000,
    callbacks: Dict[str, Callable[[BaseModel], None]],
    msg_types: Dict[str, Type[BaseModel]],
    is_shutting_down_f: Callable[[], bool],
    **kwargs,
):
    """todo: write docs"""
    logger.info(f"aiokafka_consumer_loop() starting..")
    consumer = AIOKafkaConsumer(
        bootstrap_servers=bootstrap_servers,
        auto_offset_reset=auto_offset_reset,
        max_poll_records=max_poll_records,
    )
    logger.info("aiokafka_consumer_loop(): Consumer created.")

    await consumer.start()
    logger.info("aiokafka_consumer_loop(): Consumer started.")
    consumer.subscribe(topics)
    logger.info("aiokafka_consumer_loop(): Consumer subscribed.")

    try:
        await _aiokafka_consumer_loop(
            consumer=consumer,
            max_buffer_size=max_buffer_size,
            timeout_ms=timeout_ms,
            callbacks=callbacks,
            msg_types=msg_types,
            is_shutting_down_f=is_shutting_down_f,
        )
    finally:
        await consumer.stop()
        logger.info(f"aiokafka_consumer_loop(): Consumer stopped.")
        logger.info(f"aiokafka_consumer_loop() finished.")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/002_ProducerLoop.ipynb.

# %% auto 0
__all__ = ['logger']

# %% ../../nbs/002_ProducerLoop.ipynb 1
from .logger import get_logger

# %% ../../nbs/002_ProducerLoop.ipynb 5
logger = get_logger(__name__)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/fast-kafka-api',
                'doc_host': 'https://airtai.github.io',
                'git_url': 'https://github.com/airtai/fast-kafka-api',
                'lib_path': 'fast_kafka_api'},
  'syms': { 'fast_kafka_api._components.aiokafka_consumer_loop': { 'fast_kafka_api._components.aiokafka_consumer_loop._aiokafka_consumer_loop': ( 'consumerloop.html#_aiokafka_consumer_loop',
                                                                                                                                                  'fast_kafka_api/_components/aiokafka_consumer_loop.py'),
                                                                   'fast_kafka_api._components.aiokafka_consumer_loop.aiokafka_consumer_loop': ( 'consumerloop.html#aiokafka_consumer_loop',
                                                                                                                                                 'fast_kafka_api/_components/aiokafka_consumer_loop.py'),
                                                                   'fast_kafka_api._components.aiokafka_consumer_loop.process_message_callback': ( 'consumerloop.html#process_message_callback',
                                                                                                                                                   'fast_kafka_api/_components/aiokafka_consumer_loop.py'),
                                                                   'fast_kafka_api._components.aiokafka_consumer_loop.process_msgs': ( 'consumerloop.html#process_msgs',
                                                                                                                                       'fast_kafka_api/_components/aiokafka_consumer_loop.py')},
            'fast_kafka_api._components.aiokafka_loop': {},
            'fast_kafka_api._components.aiokafka_producer_loop': { 'fast_kafka_api._components.aiokafka_producer_loop.aiokafka_producer_loop': ( 'producerloop.html#aiokafka_producer_loop',
                                                                                                                                                 'fast_kafka_api/_components/aiokafka_producer_loop.py')},
            'fast_kafka_api._components.aiokafka_producer_manager': { 'fast_kafka_api._components.aiokafka_producer_manager.AIOKafkaProducerManager': ( 'producermanager.html#aiokafkaproducermanager',
                                                                                                                                                        'fast_kafka_api/_components/aiokafka_producer_manager.py'),
                                                                      'fast_kafka_api._components.aiokafka_producer_manager.AIOKafkaProducerManager.__init__': ( 'producermanager.html#aiokafkaproducermanager.__init__',
                                                                                                                                                                 'fast_kafka_api/_components/aiokafka_producer_manager.py'),
                                                                      'fast_kafka_api._components.aiokafka_producer_manager.AIOKafkaProducerManager.send': ( 'producermanager.html#aiokafkaproducermanager.send',
                                                                                                                                                             'fast_kafka_api/_components/aiokafka_producer_manager.py'),
                                                                      'fast_kafka_api._components.aiokafka_producer_manager.AIOKafkaProducerManager.start': ( 'producermanager.html#aiokafkaproducermanager.start',
                                                                                                                                                              'fast_kafka_api/_components/aiokafka_producer_manager.py'),
                                                                      'fast_kafka_api._components.aiokafka_producer_manager.AIOKafkaProducerManager.stop': ( 'producermanager.html#aiokafkaproducermanager.stop',
                                                                                                                                                             'fast_kafka_api/_components/aiokafka_producer_manager.py'),
                                                                      'fast_kafka_api._components.aiokafka_producer_manager._aiokafka_producer_manager': ( 'producermanager.html#_aiokafka_producer_manager',
                                                                                                                                                           'fast_kafka_api/_components/aiokafka_producer_manager.py')},
            'fast_kafka_api._components.asyncapi': { 'fast_kafka_api._components.asyncapi.APIKeyLocation': ( 'asyncapi.html#apikeylocation',
                                                                                                             'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.ContactInfo': ( 'asyncapi.html#contactinfo',
                                                                                                          'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.KafkaBroker': ( 'asyncapi.html#kafkabroker',
                                                                                                          'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.KafkaBroker.dict': ( 'asyncapi.html#kafkabroker.dict',
                                                                                                               'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.KafkaBroker.json': ( 'asyncapi.html#kafkabroker.json',
                                                                                                               'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.KafkaBrokers': ( 'asyncapi.html#kafkabrokers',
                                                                                                           'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.KafkaMessage': ( 'asyncapi.html#kafkamessage',
                                                                                                           'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.KafkaMessage.Config': ( 'asyncapi.html#kafkamessage.config',
                                                                                                                  'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.KafkaServiceInfo': ( 'asyncapi.html#kafkaserviceinfo',
                                                                                                               'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.SecuritySchema': ( 'asyncapi.html#securityschema',
                                                                                                             'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.SecuritySchema.__init__': ( 'asyncapi.html#securityschema.__init__',
                                                                                                                      'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.SecuritySchema.dict': ( 'asyncapi.html#securityschema.dict',
                                                                                                                  'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.SecuritySchema.json': ( 'asyncapi.html#securityschema.json',
                                                                                                                  'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.SecurityType': ( 'asyncapi.html#securitytype',
                                                                                                           'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._add_example_to_msg_definitions': ( 'asyncapi.html#_add_example_to_msg_definitions',
                                                                                                                              'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_asyncapi_schema': ( 'asyncapi.html#_get_asyncapi_schema',
                                                                                                                   'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_channels_schema': ( 'asyncapi.html#_get_channels_schema',
                                                                                                                   'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_components_schema': ( 'asyncapi.html#_get_components_schema',
                                                                                                                     'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_example': ( 'asyncapi.html#_get_example',
                                                                                                           'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_kafka_msg_classes': ( 'asyncapi.html#_get_kafka_msg_classes',
                                                                                                                     'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_kafka_msg_definitions': ( 'asyncapi.html#_get_kafka_msg_definitions',
                                                                                                                         'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_msg_cls_for_consumer': ( 'asyncapi.html#_get_msg_cls_for_consumer',
                                                                                                                        'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_msg_cls_for_producer': ( 'asyncapi.html#_get_msg_cls_for_producer',
                                                                                                                        'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_msg_definitions_with_examples': ( 'asyncapi.html#_get_msg_definitions_with_examples',
                                                                                                                                 'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_security_schemes': ( 'asyncapi.html#_get_security_schemes',
                                                                                                                    'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_servers_schema': ( 'asyncapi.html#_get_servers_schema',
                                                                                                                  'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi._get_topic_dict': ( 'asyncapi.html#_get_topic_dict',
                                                                                                              'fast_kafka_api/_components/asyncapi.py'),
                                                     'fast_kafka_api._components.asyncapi.export_async_spec': ( 'asyncapi.html#export_async_spec',
                                                                                                                'fast_kafka_api/_components/asyncapi.py')},
            'fast_kafka_api._components.logger': { 'fast_kafka_api._components.logger.get_default_logger_configuration': ( 'logger.html#get_default_logger_configuration',
                                                                                                                           'fast_kafka_api/_components/logger.py'),
                                                   'fast_kafka_api._components.logger.get_logger': ( 'logger.html#get_logger',
                                                                                                     'fast_kafka_api/_components/logger.py'),
                                                   'fast_kafka_api._components.logger.set_level': ( 'logger.html#set_level',
                                                                                                    'fast_kafka_api/_components/logger.py'),
                                                   'fast_kafka_api._components.logger.supress_timestamps': ( 'logger.html#supress_timestamps',
                                                                                                             'fast_kafka_api/_components/logger.py')},
            'fast_kafka_api.application': { 'fast_kafka_api.application.FastKafkaAPI': ( 'fastkafkaapi.html#fastkafkaapi',
                                                                                         'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI.__init__': ( 'fastkafkaapi.html#fastkafkaapi.__init__',
                                                                                                  'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI._on_shutdown': ( 'fastkafkaapi.html#fastkafkaapi._on_shutdown',
                                                                                                      'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI._on_startup': ( 'fastkafkaapi.html#fastkafkaapi._on_startup',
                                                                                                     'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI._populate_consumers': ( 'fastkafkaapi.html#fastkafkaapi._populate_consumers',
                                                                                                             'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI._populate_producers': ( 'fastkafkaapi.html#fastkafkaapi._populate_producers',
                                                                                                             'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI._shutdown_consumers': ( 'fastkafkaapi.html#fastkafkaapi._shutdown_consumers',
                                                                                                             'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI._shutdown_producers': ( 'fastkafkaapi.html#fastkafkaapi._shutdown_producers',
                                                                                                             'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI.consumes': ( 'fastkafkaapi.html#fastkafkaapi.consumes',
                                                                                                  'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.FastKafkaAPI.produces': ( 'fastkafkaapi.html#fastkafkaapi.produces',
                                                                                                  'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application._create_producer': ( 'fastkafkaapi.html#_create_producer',
                                                                                             'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application._get_topic_name': ( 'fastkafkaapi.html#_get_topic_name',
                                                                                            'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application._to_json_utf8': ( 'fastkafkaapi.html#_to_json_utf8',
                                                                                          'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.filter_using_signature': ( 'fastkafkaapi.html#filter_using_signature',
                                                                                                   'fast_kafka_api/application.py'),
                                            'fast_kafka_api.application.produce_decorator': ( 'fastkafkaapi.html#produce_decorator',
                                                                                              'fast_kafka_api/application.py')},
            'fast_kafka_api.testing': { 'fast_kafka_api.testing.create_and_fill_testing_topic': ( 'test_utils.html#create_and_fill_testing_topic',
                                                                                                  'fast_kafka_api/testing.py'),
                                        'fast_kafka_api.testing.create_missing_topics': ( 'test_utils.html#create_missing_topics',
                                                                                          'fast_kafka_api/testing.py'),
                                        'fast_kafka_api.testing.create_testing_topic': ( 'test_utils.html#create_testing_topic',
                                                                                         'fast_kafka_api/testing.py'),
                                        'fast_kafka_api.testing.mock_AIOKafkaProducer_send': ( 'test_utils.html#mock_aiokafkaproducer_send',
                                                                                               'fast_kafka_api/testing.py'),
                                        'fast_kafka_api.testing.nb_safe_seed': ( 'test_utils.html#nb_safe_seed',
                                                                                 'fast_kafka_api/testing.py'),
                                        'fast_kafka_api.testing.true_after': ('test_utils.html#true_after', 'fast_kafka_api/testing.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/999_Test_Utils.ipynb.

# %% auto 0
__all__ = ['logger', 'kafka_server_url', 'kafka_server_port', 'kafka_config', 'true_after', 'create_missing_topics',
           'create_testing_topic', 'create_and_fill_testing_topic', 'nb_safe_seed', 'mock_AIOKafkaProducer_send']

# %% ../nbs/999_Test_Utils.ipynb 1
from typing import List, Dict, Any, Optional, Callable, Tuple, Generator
from os import environ
from contextlib import contextmanager, asynccontextmanager
import random
from datetime import datetime, timedelta
import time
import asyncio
import hashlib

import unittest

from confluent_kafka.admin import AdminClient, NewTopic
from aiokafka import AIOKafkaProducer, AIOKafkaConsumer

from ._components.logger import get_logger

# %% ../nbs/999_Test_Utils.ipynb 3
logger = get_logger(__name__)

# %% ../nbs/999_Test_Utils.ipynb 5
kafka_server_url = (
    environ["KAFKA_HOSTNAME"] if "KAFKA_HOSTNAME" in environ else "localhost"
)
kafka_server_port = environ["KAFKA_PORT"] if "KAFKA_PORT" in environ else "9092"

kafka_config = {
    "bootstrap.servers": f"{kafka_server_url}:{kafka_server_port}",
    # "group.id": f"{kafka_server_url}:{kafka_server_port}_group"
}

# %% ../nbs/999_Test_Utils.ipynb 6
def true_after(seconds: float):
    """Function returning True after a given number of seconds"""
    t = datetime.now()

    def _true_after(seconds=seconds, t=t):
        return (datetime.now() - t) > timedelta(seconds=seconds)

    return _true_after

# %% ../nbs/999_Test_Utils.ipynb 8
## TODO: Check if replication num is <= of number of brokers
## TODO: Add tests for:
#             - Replication factor (less than and greater than number of brokers)
#             - Num partitions


def create_missing_topics(
    admin: AdminClient,
    topic_names: List[str],
    *,
    num_partitions: Optional[int] = None,
    replication_factor: Optional[int] = None,
    **kwargs,
) -> None:
    if not replication_factor:
        replication_factor = len(admin.list_topics().brokers)
    if not num_partitions:
        num_partitions = replication_factor
    existing_topics = list(admin.list_topics().topics.keys())
    logger.debug(
        f"create_missing_topics({topic_names}): existing_topics={existing_topics}, num_partitions={num_partitions}, replication_factor={replication_factor}"
    )
    new_topics = [
        NewTopic(
            topic,
            num_partitions=num_partitions,
            replication_factor=replication_factor,
            **kwargs,
        )
        for topic in topic_names
        if topic not in existing_topics
    ]
    if len(new_topics):
        logger.info(f"create_missing_topics({topic_names}): new_topics = {new_topics}")
        fs = admin.create_topics(new_topics)
        while not set(topic_names).issubset(set(admin.list_topics().topics.keys())):
            time.sleep(1)

# %% ../nbs/999_Test_Utils.ipynb 10
@contextmanager
def create_testing_topic(
    kafka_config: Dict[str, Any], topic_prefix: str, seed: Optional[int] = None
) -> Generator[str, None, None]:
    # create random topic name
    random.seed(seed)
    topic = topic_prefix + str(random.randint(0, 10**10)).zfill(3)

    # delete topic if it already exists
    admin = AdminClient(kafka_config)
    existing_topics = admin.list_topics().topics.keys()
    if topic in existing_topics:
        logger.warning(f"topic {topic} exists, deleting it...")
        fs = admin.delete_topics(topics=[topic])
        results = {k: f.result() for k, f in fs.items()}
        while topic in admin.list_topics().topics.keys():
            time.sleep(1)
    try:
        # create topic if needed
        create_missing_topics(admin, [topic])
        while topic not in admin.list_topics().topics.keys():
            time.sleep(1)
        yield topic

    finally:
        pass
        # cleanup if needed again
        fs = admin.delete_topics(topics=[topic])
        while topic in admin.list_topics().topics.keys():
            time.sleep(1)

# %% ../nbs/999_Test_Utils.ipynb 12
@asynccontextmanager
async def create_and_fill_testing_topic(
    msgs: List[bytes], kafka_config: Dict[str, str] = kafka_config, *, seed: int
) -> Generator[str, None, None]:

    with create_testing_topic(kafka_config, "my_topic_", seed=seed) as topic:

        producer = AIOKafkaProducer(bootstrap_servers=kafka_config["bootstrap.servers"])
        logger.info(f"Producer {producer} created.")

        await producer.start()
        logger.info(f"Producer {producer} started.")
        try:
            fx = [
                producer.send(
                    topic,
                    msg,
                    key=f"{i % 17}".encode("utf-8"),
                )
                for i, msg in enumerate(msgs)
            ]
            await producer.flush()
            sent_msgs = [await f for f in fx]
            msg_statuses = [await s for s in sent_msgs]
            logger.info(f"Sent messages: len(sent_msgs)={len(sent_msgs)}")

            yield topic
        finally:
            await producer.stop()
            logger.info(f"Producer {producer} stoped.")

# %% ../nbs/999_Test_Utils.ipynb 15
def nb_safe_seed(s: str) -> Callable[[int], int]:
    """Gets a unique seed function for a notebook

    Params:
        s: name of the notebook used to initialize the seed function

    Returns:
        A unique seed function
    """
    init_seed = int(hashlib.sha1(s.encode("utf-8")).hexdigest(), 16) % (10**8)

    def _get_seed(x: int = 0, *, init_seed: int = init_seed) -> int:
        return init_seed + x

    return _get_seed

# %% ../nbs/999_Test_Utils.ipynb 17
@contextmanager
def mock_AIOKafkaProducer_send():
    """Mocks **send** method of **AIOKafkaProducer**"""
    with unittest.mock.patch("__main__.AIOKafkaProducer.send") as mock:

        async def _f():
            pass

        mock.return_value = asyncio.create_task(_f())

        yield mock



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/002_ProducerLoop.ipynb.

# %% auto 0
__all__ = ['logger', 'aiokafka_producer_loop']

# %% ../../nbs/002_ProducerLoop.ipynb 1
from .logger import get_logger

# %% ../../nbs/002_ProducerLoop.ipynb 5
logger = get_logger(__name__)

# %% ../../nbs/002_ProducerLoop.ipynb 15
async def aiokafka_producer_loop():
    pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/000_FastKafkaAPI.ipynb.

# %% auto 0
__all__ = ['logger', 'FastKafkaAPI', 'produce_decorator', 'filter_using_signature']

# %% ../nbs/000_FastKafkaAPI.ipynb 1
from typing import *
from typing import get_type_hints

from enum import Enum
from pathlib import Path
import json
import yaml
from copy import deepcopy
from os import environ
from datetime import datetime, timedelta
import tempfile
from contextlib import contextmanager, asynccontextmanager
import time
from inspect import signature
import functools

from fastcore.foundation import patch

import anyio
import asyncio
from asyncio import iscoroutinefunction  # do not use the version from inspect
import httpx
from fastapi import FastAPI
from fastapi import status, Depends, HTTPException, Request, Response
from fastapi.openapi.docs import get_swagger_ui_html, get_redoc_html
from fastapi.openapi.utils import get_openapi
from fastapi.responses import FileResponse, RedirectResponse
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel
from pydantic import Field, HttpUrl, EmailStr, PositiveInt
from pydantic.schema import schema
from pydantic.json import timedelta_isoformat
from aiokafka import AIOKafkaProducer, AIOKafkaConsumer

import confluent_kafka
from confluent_kafka import Producer
from confluent_kafka.admin import AdminClient, NewTopic
from confluent_kafka import Message, KafkaError
import asyncer

import fast_kafka_api._components.logger

fast_kafka_api._components.logger.should_supress_timestamps = True

import fast_kafka_api
from ._components.aiokafka_consumer_loop import aiokafka_consumer_loop
from ._components.aiokafka_producer_manager import AIOKafkaProducerManager
from fast_kafka_api._components.asyncapi import (
    KafkaMessage,
    export_async_spec,
    ConsumeCallable,
    ProduceCallable,
    KafkaBroker,
    ContactInfo,
    KafkaServiceInfo,
    KafkaBrokers,
)
from ._components.logger import get_logger, supress_timestamps

# %% ../nbs/000_FastKafkaAPI.ipynb 2
logger = get_logger(__name__)

# %% ../nbs/000_FastKafkaAPI.ipynb 8
class FastKafkaAPI(FastAPI):
    def __init__(
        self,
        *,
        title: str = "FastKafkaAPI",
        kafka_config: Dict[str, Any],
        contact: Optional[Dict[str, Union[str, Any]]] = None,
        kafka_brokers: Optional[Dict[str, Any]] = None,
        root_path: Optional[Union[Path, str]] = None,
        **kwargs,
    ):
        """Combined REST and Kafka service

        Params:
            title: name of the service, used for generating documentation
            kafka_config:
            contact:
            kafka_brokers:
            root_path:
            kwargs: parameters passed to FastAPI constructor
        """
        self._kafka_config = kafka_config

        config_defaults = {
            "bootstrap_servers": "localhost:9092",
            "auto_offset_reset": "earliest",
            "max_poll_records": 100,
            "max_buffer_size": 10_000,
        }

        for key, value in config_defaults.items():
            if key not in kafka_config:
                kafka_config[key] = value

        if root_path is None:
            root_path = Path(".")
        self._root_path = Path(root_path)

        if kafka_brokers is None:
            kafka_brokers = {
                "localhost": KafkaBroker(
                    url="https://localhost",
                    description="Local (dev) Kafka broker",
                    port="9092",
                )
            }
        if contact is None:
            contact = dict(
                name="author", url="https://www.google.com", email="noreply@gmail.com"
            )

        super().__init__(title=title, contact=contact, **kwargs)

        self._consumers_store: Dict[str, Tuple[ConsumeCallable, Dict[str, Any]]] = {}

        self._producers_store: Dict[
            str, Tuple[ProduceCallable, AIOKafkaProducer, Dict[str, Any]]
        ] = {}
        self._on_error_topic: Optional[str] = None

        contact_info = ContactInfo(**contact)  # type: ignore
        self._kafka_service_info = KafkaServiceInfo(
            title=self.title,
            version=self.version,
            description=self.description,
            contact=contact_info,
        )

        self._kafka_brokers = KafkaBrokers(brokers=kafka_brokers)

        self._asyncapi_path = self._root_path / "asyncapi"
        (self._asyncapi_path / "docs").mkdir(exist_ok=True, parents=True)
        (self._asyncapi_path / "spec").mkdir(exist_ok=True, parents=True)
        self.mount(
            "/asyncapi",
            StaticFiles(directory=self._asyncapi_path / "docs"),
            name="asyncapi",
        )

        self._is_shutting_down: bool = False
        self._kafka_consumer_tasks: List[asyncio.Task[Any]] = []
        self._kafka_producer_tasks: List[asyncio.Task[Any]] = []

        @self.get("/", include_in_schema=False)
        def redirect_root_to_asyncapi():
            return RedirectResponse("/asyncapi")

        @self.get("/asyncapi", include_in_schema=False)
        async def redirect_asyncapi_docs():
            return RedirectResponse("/asyncapi/index.html")

        @self.get("/asyncapi.yml", include_in_schema=False)
        async def download_asyncapi_yml():
            return FileResponse(self._asyncapi_path / "spec" / "asyncapi.yml")

        @self.on_event("startup")
        async def on_startup(app=self):
            await app._on_startup()

        @self.on_event("shutdown")
        async def on_shutdown(app=self):
            await app._on_shutdown()

    async def _on_startup(self) -> None:
        raise NotImplementedError

    async def _on_shutdown(self) -> None:
        raise NotImplementedError

    def consumes(
        self,
        topic: Optional[str] = None,
        *,
        prefix: str = "on_",
        **kwargs,
    ) -> ConsumeCallable:
        raise NotImplementedError

    def produces(
        self,
        topic: Optional[str] = None,
        *,
        prefix: str = "to_",
        producer: AIOKafkaProducer = None,
        **kwargs,
    ) -> ProduceCallable:
        raise NotImplementedError

# %% ../nbs/000_FastKafkaAPI.ipynb 9
def _get_topic_name(
    topic_callable: Union[ConsumeCallable, ProduceCallable], prefix: str = "on_"
) -> str:
    topic = topic_callable.__name__
    if not topic.startswith(prefix) or len(topic) <= len(prefix):
        raise ValueError(f"Function name '{topic}' must start with {prefix}")
    topic = topic[len(prefix) :]

    return topic

# %% ../nbs/000_FastKafkaAPI.ipynb 13
@patch
def consumes(
    self: FastKafkaAPI,
    topic: Optional[str] = None,
    *,
    prefix: str = "on_",
    **kwargs,
) -> ConsumeCallable:
    """Decorator registering the callback called when a message is received in a topic.

    This function decorator is also responsible for registering topics for AsyncAPI specificiation and documentation.

    Params:
        topic: Kafka topic that the consumer will subscribe to and execute the decorated function when it receives a message from the topic, default: None
            If the topic is not specified, topic name will be inferred from the decorated function name by stripping the defined prefix
        prefix: Prefix stripped from the decorated function to define a topic name if the topic argument is not passed, default: "on_"
            If the decorated function name is not prefixed with the defined prefix and topic argument is not passed, then this method will throw ValueError
        **kwargs: Keyword arguments that will be passed to AIOKafkaConsumer, used to configure the consumer

    Returns:
        A function returning the same function

    Throws:
        ValueError

    """

    def _decorator(
        on_topic: ConsumeCallable, topic: str = topic, kwargs=kwargs
    ) -> ConsumeCallable:
        if topic is None:
            topic = _get_topic_name(topic_callable=on_topic, prefix=prefix)

        self._consumers_store[topic] = (on_topic, kwargs)

        return on_topic

    return _decorator

# %% ../nbs/000_FastKafkaAPI.ipynb 15
def _to_json_utf8(o: Any) -> bytes:
    """Converts to JSON and then encodes with UTF-8"""
    if hasattr(o, "json"):
        return o.json().encode("utf-8")
    else:
        return json.dumps(o).encode("utf-8")

# %% ../nbs/000_FastKafkaAPI.ipynb 18
def produce_decorator(self: FastKafkaAPI, func: ProduceCallable, topic: str):
    @functools.wraps(func)
    async def _produce_async(*args, **kwargs):
        return_val = await func(*args, **kwargs)
        _, producer, _ = self._producers_store[topic]
        fut = await producer.send(topic, _to_json_utf8(return_val))
        msg = await fut
        return return_val

    @functools.wraps(func)
    def _produce_sync(*args, **kwargs):
        return_val = func(*args, **kwargs)
        _, producer, _ = self._producers_store[topic]
        producer.send(topic, _to_json_utf8(return_val))
        return return_val

    return _produce_async if iscoroutinefunction(func) else _produce_sync

# %% ../nbs/000_FastKafkaAPI.ipynb 20
@patch
def produces(
    self: FastKafkaAPI,
    topic: Optional[str] = None,
    *,
    prefix: str = "to_",
    producer: AIOKafkaProducer = None,
    **kwargs,
) -> ProduceCallable:
    """Decorator registering the callback called when delivery report for a produced message is received

    This function decorator is also responsible for registering topics for AsyncAPI specificiation and documentation.

    Params:
        topic: Kafka topic that the producer will send returned values from the decorated function to, default: None
            If the topic is not specified, topic name will be inferred from the decorated function name by stripping the defined prefix
        prefix: Prefix stripped from the decorated function to define a topic name if the topic argument is not passed, default: "to_"
            If the decorated function name is not prefixed with the defined prefix and topic argument is not passed, then this method will throw ValueError
        producer:
        **kwargs: Keyword arguments that will be passed to AIOKafkaProducer, used to configure the producer

    Returns:
        A function returning the same function

    Throws:
        ValueError

    """

    def _decorator(
        on_topic: ProduceCallable, topic: str = topic, kwargs=kwargs
    ) -> ProduceCallable:
        if topic is None:
            topic = _get_topic_name(topic_callable=on_topic, prefix=prefix)

        self._producers_store[topic] = (on_topic, producer, kwargs)

        return produce_decorator(self, on_topic, topic)

    return _decorator

# %% ../nbs/000_FastKafkaAPI.ipynb 24
def filter_using_signature(f: Callable, **kwargs) -> Dict[str, Any]:
    param_names = list(signature(f).parameters.keys())
    return {k: v for k, v in kwargs.items() if k in param_names}

# %% ../nbs/000_FastKafkaAPI.ipynb 26
@patch
def _populate_consumers(
    self: FastKafkaAPI,
    is_shutting_down_f: Callable[[], bool],
) -> None:
    default_config: Dict[str, Any] = filter_using_signature(
        AIOKafkaConsumer, **self._kafka_config
    )
    self._kafka_consumer_tasks = [
        asyncio.create_task(
            aiokafka_consumer_loop(
                topics=[topic],
                callbacks={topic: consumer},
                msg_types={topic: signature(consumer).parameters["msg"].annotation},
                is_shutting_down_f=is_shutting_down_f,
                **{**default_config, **override_config}
            )
        )
        for topic, (consumer, override_config) in self._consumers_store.items()
    ]


@patch
async def _shutdown_consumers(
    self: FastKafkaAPI,
) -> None:
    if self._kafka_consumer_tasks:
        await asyncio.wait(self._kafka_consumer_tasks)

# %% ../nbs/000_FastKafkaAPI.ipynb 28
# TODO: Add passing of vars
async def _create_producer(
    *,
    callback: ProduceCallable,
    producer: Optional[AIOKafkaProducer],
    default_config: Dict[str, Any],
    override_config: Dict[str, Any],
    producers_list: List[Union[AIOKafkaProducer, AIOKafkaProducerManager]]
) -> Union[AIOKafkaProducer, AIOKafkaProducerManager]:
    """Creates a producer

    Args:
        callback: A callback function that is called when the producer is ready.
        producer: An existing producer to use.
        default_config: A dictionary of default configuration values.
        override_config: A dictionary of configuration values to override.
        producers_list: A list of producers to add the new producer to.

    Returns:
        A producer.
    """

    if producer is None:
        config = {
            **filter_using_signature(AIOKafkaProducer, **default_config),
            **override_config,
        }
        producer = AIOKafkaProducer(**config)

    if not iscoroutinefunction(callback):
        producer = AIOKafkaProducerManager(producer)

    await producer.start()

    producers_list.append(producer)

    return producer


@patch
async def _populate_producers(self: FastKafkaAPI) -> None:
    """Populates the producers for the FastKafkaAPI instance.

    Args:
        self: The FastKafkaAPI instance.

    Returns:
        None.

    Raises:
        None.
    """
    default_config: Dict[str, Any] = self._kafka_config
    self._producers_list = []
    self._producers_store = {
        topic: (
            callback,
            await _create_producer(
                callback=callback,
                producer=producer,
                default_config=default_config,
                override_config=override_config,
                producers_list=self._producers_list,
            ),
            override_config,
        )
        for topic, (
            callback,
            producer,
            override_config,
        ) in self._producers_store.items()
    }


@patch
async def _shutdown_producers(self: FastKafkaAPI) -> None:
    [await producer.stop() for producer in self._producers_list[::-1]]

# %% ../nbs/000_FastKafkaAPI.ipynb 30
@patch
async def _on_startup(self: FastKafkaAPI) -> None:
    export_async_spec(
        consumers={topic: callback for topic, (callback, _) in self._consumers_store.items()},  # type: ignore
        producers={topic: callback for topic, (callback, _, _) in self._producers_store.items()},  # type: ignore
        kafka_brokers=self._kafka_brokers,
        kafka_service_info=self._kafka_service_info,
        asyncapi_path=self._asyncapi_path,
    )

    self._is_shutting_down = False

    def is_shutting_down_f(self: FastKafkaAPI = self) -> bool:
        return self._is_shutting_down

    self._populate_consumers(is_shutting_down_f)
    await self._populate_producers()


@patch
async def _on_shutdown(self: FastKafkaAPI) -> None:
    self._is_shutting_down = True

    await self._shutdown_consumers()
    await self._shutdown_producers()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Logger.ipynb.

# %% auto 0
__all__ = ['should_supress_timestamps', 'logger_spaces_added', 'supress_timestamps', 'get_default_logger_configuration',
           'get_logger', 'set_level']

# %% ../../nbs/Logger.ipynb 2
from typing import *
import logging

# %% ../../nbs/Logger.ipynb 3
import logging.config
from contextlib import contextmanager, redirect_stdout, redirect_stderr
from io import StringIO
import traceback

# %% ../../nbs/Logger.ipynb 5
# Logger Levels
# CRITICAL = 50
# ERROR = 40
# WARNING = 30
# INFO = 20
# DEBUG = 10
# NOTSET = 0

should_supress_timestamps: bool = False


def supress_timestamps(flag: bool = True):
    """Supress logger timestamp

    Args:
        flag: If not set, then the default value **True** will be used to supress the timestamp
            from the logger messages
    """
    global should_supress_timestamps
    should_supress_timestamps = flag


def get_default_logger_configuration(level: int = logging.INFO) -> Dict[str, Any]:
    """Return the common configurations for the logger

    Args:
        level: Logger level to set

    Returns:
        A dict with default logger configuration

    """
    global should_supress_timestamps

    if should_supress_timestamps:
        FORMAT = "[%(levelname)s] %(name)s: %(message)s"
    else:
        FORMAT = "%(asctime)s.%(msecs)03d [%(levelname)s] %(name)s: %(message)s"

    DATE_FMT = "%y-%m-%d %H:%M:%S"

    LOGGING_CONFIG = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {"format": FORMAT, "datefmt": DATE_FMT},
        },
        "handlers": {
            "default": {
                "level": level,
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",  # Default is stderr
            },
        },
        "loggers": {
            "": {"handlers": ["default"], "level": level},  # root logger
        },
    }
    return LOGGING_CONFIG

# %% ../../nbs/Logger.ipynb 9
logger_spaces_added: List[str] = []


def get_logger(
    name: str, *, level: int = logging.INFO, add_spaces: bool = True
) -> logging.Logger:
    """Return the logger class with default logging configuration.

    Args:
        name: Pass the __name__ variable as name while calling
        level: Used to configure logging, default value `logging.INFO` logs
            info messages and up.
        add_spaces:

    Returns:
        The logging.Logger class with default/custom logging configuration

    """
    config = get_default_logger_configuration(level=level)
    logging.config.dictConfig(config)

    logger = logging.getLogger(name)
    #     stack_size = len(traceback.extract_stack())
    #     def add_spaces_f(f):
    #         def f_with_spaces(msg, *args, **kwargs):
    #             cur_stack_size = len(traceback.extract_stack())
    #             msg = " "*(cur_stack_size-stack_size)*2 + msg
    #             return f(msg, *args, **kwargs)
    #         return f_with_spaces

    #     if name not in logger_spaces_added and add_spaces:
    #         logger.debug = add_spaces_f(logger.debug) # type: ignore
    #         logger.info = add_spaces_f(logger.info) # type: ignore
    #         logger.warning = add_spaces_f(logger.warning) # type: ignore
    #         logger.error = add_spaces_f(logger.error) # type: ignore
    #         logger.critical = add_spaces_f(logger.critical) # type: ignore
    #         logger.exception = add_spaces_f(logger.exception) # type: ignore

    #         logger_spaces_added.append(name)

    return logger

# %% ../../nbs/Logger.ipynb 15
def set_level(level: int):
    """Set logger level

    Args:
        level: Logger level to set
    """

    # Getting all loggers that has either airt or __main__ in the name
    loggers = [
        logging.getLogger(name)
        for name in logging.root.manager.loggerDict
        if ("airt" in name) or ("__main__" in name)
    ]

    for logger in loggers:
        logger.setLevel(level)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/099_Test_Service.ipynb.

# %% auto 0
__all__ = ['logger', 'ModelType', 'ModelTrainingRequest', 'EventData', 'RealtimeData', 'TrainingDataStatus',
           'TrainingModelStatus', 'ModelMetrics', 'Prediction', 'create_ws_server']

# %% ../nbs/099_Test_Service.ipynb 1
from typing import *

import os
import json
import yaml
from pathlib import Path
from copy import deepcopy
from datetime import datetime
from os import environ
from enum import Enum
import httpx

from confluent_kafka import Producer, Consumer
from fastapi import status, Depends, HTTPException, Request, Response
from fastapi.openapi.docs import get_swagger_ui_html, get_redoc_html
from fastapi.openapi.utils import get_openapi
from fastapi.responses import FileResponse, RedirectResponse
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from fastapi.staticfiles import StaticFiles
from pydantic import validator, BaseModel, Field, HttpUrl, EmailStr, NonNegativeInt

from fast_kafka_api.application import (
    FastKafkaAPI,
    KafkaMessage,
    ConsumeCallable,
    ProduceCallable,
)

from ._components.logger import get_logger

# %% ../nbs/099_Test_Service.ipynb 3
logger = get_logger(__name__)

# %% ../nbs/099_Test_Service.ipynb 5
class ModelType(str, Enum):
    churn = "churn"
    propensity_to_buy = "propensity_to_buy"


class ModelTrainingRequest(BaseModel):
    AccountId: NonNegativeInt = Field(
        ..., example=202020, description="ID of an account"
    )
    ModelName: ModelType = Field(..., example="churn", description="ID of an account")
    total_no_of_records: NonNegativeInt = Field(
        ...,
        example=1_000_000,
        description="total number of records (rows) to be ingested",
    )


class EventData(BaseModel):
    """
    A sequence of events for a fixed account_id
    """

    AccountId: NonNegativeInt = Field(
        ..., example=202020, description="ID of an account"
    )
    Application: Optional[str] = Field(
        None,
        example="DriverApp",
        description="Name of the application in case there is more than one for the AccountId",
    )
    DefinitionId: str = Field(
        ...,
        example="appLaunch",
        description="name of the event",
        min_length=1,
    )
    OccurredTime: datetime = Field(
        ...,
        example="2021-03-28T00:34:08",
        description="local time of the event",
    )
    OccurredTimeTicks: NonNegativeInt = Field(
        ...,
        example=1616891648496,
        description="local time of the event as the number of ticks",
    )
    PersonId: NonNegativeInt = Field(
        ..., example=12345678, description="ID of a person"
    )


class RealtimeData(BaseModel):
    event_data: EventData = Field(
        ...,
        example=dict(
            AccountId=202020,
            Application="DriverApp",
            DefinitionId="appLaunch",
            OccurredTime="2021-03-28T00:34:08",
            OccurredTimeTicks=1616891648496,
            PersonId=12345678,
        ),
        description="realtime event data",
    )
    make_prediction: bool = Field(
        ..., example=True, description="trigger prediction message in prediction topic"
    )


class TrainingDataStatus(BaseModel):
    AccountId: NonNegativeInt = Field(
        ..., example=202020, description="ID of an account"
    )
    no_of_records: NonNegativeInt = Field(
        ...,
        example=12_345,
        description="number of records (rows) ingested",
    )
    total_no_of_records: NonNegativeInt = Field(
        ...,
        example=1_000_000,
        description="total number of records (rows) to be ingested",
    )


class TrainingModelStatus(BaseModel):
    AccountId: NonNegativeInt = Field(
        ..., example=202020, description="ID of an account"
    )
    current_step: NonNegativeInt = Field(
        ...,
        example=0,
        description="number of records (rows) ingested",
    )
    current_step_percentage: float = Field(
        ...,
        example=0.21,
        description="the percentage of the current step completed",
    )
    total_no_of_steps: NonNegativeInt = Field(
        ...,
        example=1_000_000,
        description="total number of steps for training the model",
    )


class ModelMetrics(BaseModel):
    """The standard metrics for classification models.

    The most important metrics is AUC for unbalanced classes such as churn. Metrics such as
    accuracy are not very useful since they are easily maximized by outputting the most common
    class all the time.
    """

    AccountId: NonNegativeInt = Field(
        ..., example=202020, description="ID of an account"
    )
    Application: Optional[str] = Field(
        None,
        example="DriverApp",
        description="Name of the application in case there is more than one for the AccountId",
    )
    timestamp: datetime = Field(
        ...,
        example="2021-03-28T00:34:08",
        description="UTC time when the model was trained",
    )
    model_type: ModelType = Field(
        ...,
        example="churn",
        description="Name of the model used (churn, propensity to buy)",
    )
    auc: float = Field(
        ..., example=0.91, description="Area under ROC curve", ge=0.0, le=1.0
    )
    f1: float = Field(..., example=0.89, description="F-1 score", ge=0.0, le=1.0)
    precission: float = Field(
        ..., example=0.84, description="precission", ge=0.0, le=1.0
    )
    recall: float = Field(..., example=0.82, description="recall", ge=0.0, le=1.0)
    accuracy: float = Field(..., example=0.82, description="accuracy", ge=0.0, le=1.0)


class Prediction(BaseModel):
    AccountId: NonNegativeInt = Field(
        ..., example=202020, description="ID of an account"
    )
    Application: Optional[str] = Field(
        None,
        example="DriverApp",
        description="Name of the application in case there is more than one for the AccountId",
    )
    PersonId: NonNegativeInt = Field(
        ..., example=12345678, description="ID of a person"
    )
    prediction_time: datetime = Field(
        ...,
        example="2021-03-28T00:34:08",
        description="UTC time of prediction",
    )
    model_type: ModelType = Field(
        ...,
        example="churn",
        description="Name of the model used (churn, propensity to buy)",
    )
    score: float = Field(
        ...,
        example=0.4321,
        description="Prediction score (e.g. the probability of churn in the next 28 days)",
        ge=0.0,
        le=1.0,
    )

# %% ../nbs/099_Test_Service.ipynb 6
_total_no_of_records = 0
_no_of_records_received = 0


def create_ws_server(assets_path: Path = Path("./assets")) -> FastKafkaAPI:
    title = "Example for FastKafkaAPI"
    description = "A simple example on how to use FastKafkaAPI"
    version = "0.0.1"
    openapi_url = "/openapi.json"
    favicon_url = "/assets/images/favicon.ico"

    contact = dict(name="airt.ai", url="https://airt.ai", email="info@airt.ai")

    kafka_brokers = {
        "localhost": {
            "url": "kafka",
            "description": "local development kafka",
            "port": 9092,
        },
        "staging": {
            "url": "kafka.staging.acme.com",
            "description": "staging kafka",
            "port": 9092,
            "protocol": "kafka-secure",
            "security": {"type": "plain"},
        },
        "production": {
            "url": "kafka.infobip.acme.com",
            "description": "production kafka",
            "port": 9092,
            "protocol": "kafka-secure",
            "security": {"type": "plain"},
        },
    }

    kafka_server_url = environ["KAFKA_HOSTNAME"]
    kafka_server_port = environ["KAFKA_PORT"]
    kafka_config = {
        "bootstrap_servers": f"{kafka_server_url}:{kafka_server_port}",
        "group_id": f"{kafka_server_url}:{kafka_server_port}_group",
        "auto_offset_reset": "earliest",
    }
    if "KAFKA_API_KEY" in environ:
        kafka_config = {
            **kafka_config,
            **{
                "security_protocol": "SASL_SSL",
                "sasl_mechanisms": "PLAIN",
                "sasl_username": environ["KAFKA_API_KEY"],
                "sasl_password": environ["KAFKA_API_SECRET"],
            },
        }

    app = FastKafkaAPI(
        title=title,
        contact=contact,
        kafka_brokers=kafka_brokers,
        kafka_config=kafka_config,
        description=description,
        version=version,
        docs_url=None,
        redoc_url=None,
    )

    @app.get("/docs", include_in_schema=False)
    def overridden_swagger():
        return get_swagger_ui_html(
            openapi_url=openapi_url,
            title=title,
            swagger_favicon_url=favicon_url,
        )

    @app.get("/redoc", include_in_schema=False)
    def overridden_redoc():
        return get_redoc_html(
            openapi_url=openapi_url,
            title=title,
            redoc_favicon_url=favicon_url,
        )

    @app.post("/from_kafka_start")
    async def from_kafka_start(training_request: ModelTrainingRequest):
        global _total_no_of_records
        global _no_of_records_received

        _total_no_of_records = training_request.total_no_of_records
        _no_of_records_received = 0

    @app.get("/from_kafka_end")
    async def from_kafka_end():
        pass

    @app.consumes()  # type: ignore
    async def on_training_data(msg: EventData):
        # ToDo: this is not showing up in logs
        logger.debug(f"msg={msg}")
        global _total_no_of_records
        global _no_of_records_received
        _no_of_records_received = _no_of_records_received + 1

        if _no_of_records_received % 100 == 0:
            training_data_status = TrainingDataStatus(
                AccountId=EventData.AccountId,
                no_of_records=_no_of_records_received,
                total_no_of_records=_total_no_of_records,
            )
            app.produce("training_data_status", training_data_status)

    @app.consumes()  # type: ignore
    async def on_realitime_data(msg: RealtimeData):
        pass

    @app.produces()  # type: ignore
    async def to_training_data_status(msg: TrainingDataStatus) -> TrainingDataStatus:
        logger.debug(f"on_training_data_status(msg={msg}, kafka_msg={kafka_msg})")
        return msg

    @app.produces()  # type: ignore
    async def to_training_model_status(msg: str) -> TrainingModelStatus:
        logger.debug(f"on_training_model_status(msg={msg}, kafka_msg={kafka_msg})")
        return TrainingModelStatus()

    @app.produces()  # type: ignore
    async def to_model_metrics(msg: ModelMetrics) -> ModelMetrics:
        logger.debug(f"on_training_model_status(msg={msg}, kafka_msg={kafka_msg})")
        return msg

    @app.produces()  # type: ignore
    async def to_prediction(msg: Prediction) -> Prediction:
        logger.debug(f"on_realtime_data_status(msg={msg},, kafka_msg={kafka_msg})")
        return msg

    return app
